#ifndef FILEZILLA_ENGINE_LOGGING_HEADER
#define FILEZILLA_ENGINE_LOGGING_HEADER

#include <libfilezilla/logger.hpp>

namespace logmsg {
	using namespace fz::logmsg;
	auto constexpr listing = custom1;
}

#endif
