#ifndef FILEZILLA_COMMONUI_SITE_COLOR_HEADER
#define FILEZILLA_COMMONUI_SITE_COLOR_HEADER

enum class site_colour : unsigned {
	none,
	red,
	green,
	blue,
	yellow,
	cyan,
	magenta,
	orange,
	end_of_list
};

#endif
